@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # vim: set expandtab tabstop=4 shiftwidth=4:
 
 # Copyright (c) 2020 CJ Kucera (cj@apocalyptech.com)
@@ -30,8 +30,12 @@
 
 import base64
 import struct
+import random
+import binascii
 import google.protobuf
+import google.protobuf.json_format
 from . import *
+from . import datalib
 from . import OakSave_pb2, OakShared_pb2
 
 MissionState = OakSave_pb2.MissionStatusPlayerSaveGameData.MissionState
@@ -42,20 +46,44 @@ class BL3Item(object):
     ignoring `development_save_data` entirely since it doesn't seem to
     be present in actual savegames.
 
-    No idea what `pickup_order_index` is.
+    No idea what `pickup_order_index` is, though it might just have
+    something to do with the ordering when you're picking up multiple
+    things at once (in which case it's probably only really useful for
+    things like money and ammo).
 
     All these getters/setters are rather un-Pythonic; should be using
     some decorations for that instead.  Alas!
     """
 
-    def __init__(self, protobuf):
+    def __init__(self, protobuf, datawrapper):
         self.protobuf = protobuf
+        self.datawrapper = datawrapper
+        self.serial_db = datawrapper.serial_db
+        self.name_db = datawrapper.name_db
+        self.parsed = False
+        self.can_parse = True
+        (self.decrypted_serial, self.orig_seed) = BL3Item._decrypt_serial(self.protobuf.item_serial_number)
+
+        # Attributes which get filled in when we parse
+        self._version = None
+        self._balance_bits = None
+        self._balance_idx = None
+        self._balance = None
+        self._balance_short = None
+        self._eng_name = None
+        self._invdata_bits = None
+        self._invdata_idx = None
+        self._invdata = None
+        self._manufacturer_bits = None
+        self._manufacturer_idx = None
+        self._manufacturer = None
+        self._level = None
+        self._remaining_data = None
 
     @staticmethod
-    def create(serial_number, pickup_order_idx, skin_path='', is_seen=True, is_favorite=False, is_trash=False):
+    def create(datawrapper, serial_number, pickup_order_idx, skin_path='', is_seen=True, is_favorite=False, is_trash=False):
         """
         Creates a new item with the specified serial number, pickup_order_idx, and skin_path.
-
         """
 
         # Start constructing flags
@@ -75,13 +103,290 @@ class BL3Item(object):
                 pickup_order_index=pickup_order_idx,
                 flags=flags,
                 weapon_skin_path=skin_path,
+                ), datawrapper)
+
+    @staticmethod
+    def _xor_data(data, seed):
+        """
+        Run some `data` through some XOR-based obfuscation, using the
+        specified `seed`
+        """
+
+        # If the seed is 0, we basically don't do anything (though
+        # make sure we return the same datatype as below)
+        if seed == 0:
+            return [d for d in data]
+
+        # Because our seed can be negative, we do have to do the
+        # & here, even though it might not seem to make sense to
+        # do so.
+        xor = (seed >> 5) & 0xFFFFFFFF
+        temp = []
+        for i, d in enumerate(data):
+            xor = (xor * 0x10A860C1) % 0xFFFFFFFB
+            temp.append((d ^ xor) & 0xFF)
+        return temp
+
+    @staticmethod
+    def _bogodecrypt(data, seed):
+        """
+        "Decrypts" the given item `data`, using the given `seed`.
+        """
+
+        # First run it through the xor wringer.
+        temp = BL3Item._xor_data(data, seed)
+
+        # Now rotate the data
+        steps = (seed & 0x1F) % len(data)
+        return bytearray(temp[-steps:] + temp[:-steps])
+
+    @staticmethod
+    def _bogoencrypt(data, seed):
+        """
+        "Encrypts" the given `data`, using the given `seed`
+        """
+
+        # Rotate first
+        steps = (seed & 0x1F) % len(data)
+        rotated = bytearray(data[steps:] + data[:steps])
+
+        # Then run through the xor stuff
+        return bytearray(BL3Item._xor_data(rotated, seed))
+
+    @staticmethod
+    def _decrypt_serial(serial):
+        """
+        Decrypts (really just de-obfuscates) the serial number.
+        """
+
+        # Initial byte should always be 3
+        assert(serial[0] == 3)
+
+        # Seed does need to be an unsigned int
+        orig_seed = struct.unpack('>i', serial[1:5])[0]
+
+        # Do the actual "decryption"
+        decrypted = BL3Item._bogodecrypt(serial[5:], orig_seed)
+
+        # Grab the CRC stored in the serial itself
+        orig_checksum = bytearray(decrypted[:2])
+
+        # Compute the checksum ourselves to make sure we've done
+        # everything properly
+        data_to_checksum = serial[:5] + b"\xFF\xFF" + decrypted[2:]
+        computed_crc = binascii.crc32(data_to_checksum)
+        computed_checksum = struct.pack('>H',
+                ((computed_crc >> 16) ^ computed_crc) & 0xFFFF)
+        if orig_checksum != computed_checksum:
+            raise Exception('Checksum in serial ({}) does not match computed checksum ({})'.format(
+                '0x{}'.format(''.join(f'{d:02X}' for d in orig_checksum)),
+                '0x{}'.format(''.join(f'{d:02X}' for d in computed_checksum)),
                 ))
 
-    def get_serial_number(self):
-        return self.protobuf.item_serial_number
+        # Return what we decrypted
+        return (decrypted[2:], orig_seed)
 
-    def get_serial_base64(self):
-        return 'BL3({})'.format(base64.b64encode(self.get_serial_number()).decode('latin1'))
+    @staticmethod
+    def _encrypt_serial(data, seed=None):
+        """
+        Given an unencrypted `data`, return the binary serial number for
+        the item, optionally with the given `seed`.  If `seed` is not passed in,
+        a random one will be passed in.  Use a `seed` of `0` to not apply any
+        encryption/obfuscation to the data
+        """
+
+        # Pick a random seed if one wasn't given.  Taken from the BL2 CLI editor
+        if seed is None:
+            seed = random.randrange(0x100000000) - 0x80000000
+
+        # Construct our header and find the checksum
+        header = struct.pack('>Bi', 3, seed)
+        crc32 = binascii.crc32(header + b"\xFF\xFF" + data)
+        checksum = struct.pack('>H', ((crc32 >> 16) ^ crc32) & 0xFFFF)
+
+        # Return the freshly-encrypted item
+        return header + BL3Item._bogoencrypt(checksum + data, seed)
+
+    def _get_inv_db_header_part(self, category, bits):
+        """
+        Given the category name `category`, and the ArbitraryBits object `bits`,
+        containing serial number data, return a tuple containing:
+            1) The category value
+            2) The number of bits the category takes up
+            3) The numerical index of the value
+        This relies on being run during `_parse_serial`, so that `_version` is
+        populated in our object.
+        """
+        num_bits = self.serial_db.get_num_bits(category, self._version)
+        part_idx = bits.eat(num_bits)
+        part_val = self.serial_db.get_part(category, part_idx)
+        if not part_val:
+            part_val = 'unknown'
+        return (part_val, num_bits, part_idx)
+
+    def _parse_serial(self):
+        """
+        Parse our serial number, at least up to the level.  We're not going
+        to care about actual parts in here.
+        """
+
+        if not self.can_parse:
+            return
+
+        bits = datalib.ArbitraryBits(self.decrypted_serial)
+
+        # First value should always be 128, apparently
+        assert(bits.eat(8) == 128)
+
+        # Grab the serial version and check it against the max version we know about
+        self._version = bits.eat(7)
+        if self._version > self.serial_db.max_version:
+            self.can_parse = False
+            return
+
+        # Now the rest of the data we care about.
+        (self._balance,
+                self._balance_bits,
+                self._balance_idx) = self._get_inv_db_header_part('InventoryBalanceData', bits)
+        (self._invdata,
+                self._invdata_bits,
+                self._invdata_idx) = self._get_inv_db_header_part('InventoryData', bits)
+        (self._manufacturer,
+                self._manufacturer_bits,
+                self._manufacturer_idx) = self._get_inv_db_header_part('ManufacturerData', bits)
+        self._level = bits.eat(7)
+        self._remaining_data = bits.data
+
+        # At this point, if we were planning on reading parts, we'd read eat six
+        # more bits to find the number of parts, then read in that many, using a
+        # mapping to get to the correct category.  Then another four bits tells
+        # us how many anointments there are, reading from InventoryGenericPartData
+        # after that.  After anointments, I'm not totally sure what's out there.
+
+        # Parse out a "short" balance name, for convenience's sake
+        self._balance_short = self._balance.split('.')[-1]
+
+        # If we know of an English name for this balance, use it
+        self._eng_name = self.name_db.get(self._balance_short)
+
+        # Mark down that we're parsed, now.
+        self.parsed = True
+
+    def _deparse_serial(self):
+        """
+        De-parses a serial; used after we make changes to the data that gets
+        pulled out during `_parse_serial`.  At the moment, that's only going
+        to be item level changes.  Will update the serial in the protobuf as
+        well, and set the object to trigger a re-parse if anything else needs
+        to read more.
+
+        This is all pretty inefficient -- really we should just write over the
+        bits values with the new values, in-place.  But whatever, this'll do
+        for now.
+        """
+
+        if not self.can_parse:
+            return
+
+        # Construct the new item data
+        bits = datalib.ArbitraryBits()
+        bits.append_value(128, 8)
+        bits.append_value(self._version, 7)
+        bits.append_value(self._balance_idx, self._balance_bits)
+        bits.append_value(self._invdata_idx, self._invdata_bits)
+        bits.append_value(self._manufacturer_idx, self._manufacturer_bits)
+        bits.append_value(self._level, 7)
+        bits.append_data(self._remaining_data)
+        new_data = bits.get_data()
+
+        # Encode the new serial (using seed 0; unencrypted)
+        new_serial = BL3Item._encrypt_serial(new_data, 0)
+
+        # Load in the new serial (this will set `parsed` to `False`)
+        self.set_serial_number_data(new_serial)
+
+    @property
+    def balance(self):
+        """
+        Returns the balance for this item
+        """
+        if not self.parsed:
+            self._parse_serial()
+            if not self.can_parse:
+                return None
+        return self._balance
+
+    @property
+    def balance_short(self):
+        """
+        Returns the "short" balance for this item
+        """
+        if not self.parsed:
+            self._parse_serial()
+            if not self.can_parse:
+                return None
+        return self._balance_short
+
+    @property
+    def eng_name(self):
+        """
+        Returns an English name for the balance, if possible.  Will default
+        to the "short" balance for this item if not.
+        """
+        if not self.parsed:
+            self._parse_serial()
+            if not self.can_parse:
+                return None
+        if self._eng_name:
+            return self._eng_name
+        else:
+            return self._balance_short
+
+    @property
+    def level(self):
+        """
+        Returns the level of this item
+        """
+        if not self.parsed:
+            self._parse_serial()
+            if not self.can_parse:
+                return None
+        return self._level
+
+    @level.setter
+    def level(self, value):
+        """
+        Sets a new level for the item.  This would be a super inefficient way of
+        doing it if we supported doing anything other than changing level -- we're
+        rebuilding the whole serial right now and triggering a re-parse if anything
+        decides to re-read it.  That should be sufficient for our purposes here,
+        though.
+        """
+        if not self.parsed:
+            self._parse_serial()
+            if not self.can_parse:
+                return None
+
+        # Set the level and trigger a re-encode of the serial
+        self._level = value
+        self._deparse_serial()
+
+    def get_serial_number(self, orig_seed=False):
+        """
+        Returns the binary item serial number.  If `orig_seed` is `True`, the
+        serial number will use the same seed that was used in the savegame.
+        Otherwise, it will use a seed of `0`, which will then be unencrypted.
+        """
+        return BL3Item._encrypt_serial(self.decrypted_serial, 0)
+
+    def get_serial_base64(self, orig_seed=False):
+        """
+        Returns the base64-encoded item serial number.  If `orig_seed` is
+        `True`, the serial number will use the same seed that was used in the
+        savegame.  Otherwise, it will use a seed of `0`, which will then be
+        unencrypted.
+        """
+        return 'BL3({})'.format(base64.b64encode(self.get_serial_number(orig_seed)).decode('latin1'))
 
     def get_pickup_order_idx(self):
         return self.protobuf.pickup_order_index
@@ -91,6 +396,9 @@ class BL3Item(object):
         Overwrites this item with a new one
         """
         self.protobuf.item_serial_number = new_data
+        (self.decrypted_serial, self.orig_seed) = BL3Item._decrypt_serial(self.protobuf.item_serial_number)
+        self.parsed = False
+        self.can_parse = True
 
     @staticmethod
     def decode_serial_base64(new_data):
@@ -185,6 +493,7 @@ class BL3Save(object):
 
     def __init__(self, filename, debug=False):
         self.filename = filename
+        self.datawrapper = datalib.DataWrapper()
         with open(filename, 'rb') as df:
 
             header = df.read(4)
@@ -225,7 +534,7 @@ class BL3Save(object):
                 self.custom_format_data.append((guid, entry))
             self.sg_type = self._read_str(df)
             if debug:
-                print('Savegame type: {}'.format(sg_type))
+                print('Savegame type: {}'.format(self.sg_type))
 
             # Read in the actual data
             remaining_data_len = self._read_int(df)
@@ -258,9 +567,19 @@ class BL3Save(object):
         self.save = OakSave_pb2.Character()
         self.save.ParseFromString(data)
 
+        # Some sanity checks, since this is a potentially problematic
+        # operation.
+        assert(self.save.IsInitialized())
+        assert(len(self.save.UnknownFields()) == 0)
+
+        # Not checking on byte size, in case anyone had v2 protobufs
+        # exported and were trying to import them now that we've
+        # otherwise switched to v3.
+        #assert(len(data) == self.save.ByteSize())
+
         # Do some data processing so that we can wrap things APIwise
         # First: Items
-        self.items = [BL3Item(i) for i in self.save.inventory_items]
+        self.items = [BL3Item(i, self.datawrapper) for i in self.save.inventory_items]
 
         # Next: Equip slots
         self.equipslots = {}
@@ -269,6 +588,15 @@ class BL3Save(object):
             slot = slotobj_to_slot[equip.get_obj_name()]
             self.equipslots[slot] = equip
 
+    def import_json(self, json_str):
+        """
+        Given JSON data, convert to protobuf and load it into ourselves so
+        that we can work with it.  This also sets up a few convenience vars
+        for our later use
+        """
+        message = google.protobuf.json_format.Parse(json_str, OakSave_pb2.Character())
+        self.import_protobuf(message.SerializeToString())
+
     def save_to(self, filename):
         """
         Saves ourselves to a new filename
@@ -314,6 +642,16 @@ class BL3Save(object):
         with open(filename, 'wb') as df:
             df.write(self.save.SerializeToString())
 
+    def save_json_to(self, filename):
+        """
+        Saves a JSON version of our protobuf to the specfied filename
+        """
+        with open(filename, 'w') as df:
+            df.write(google.protobuf.json_format.MessageToJson(self.save,
+                including_default_value_fields=True,
+                preserving_proto_field_name=True,
+                ))
+
     def _read_int(self, df):
         return struct.unpack('<I', df.read(4))[0]
 
@@ -840,34 +1178,72 @@ class BL3Save(object):
             elif slot == COM:
                 self.unlock_char_com_challenge()
 
-    def add_new_item(self, itemdata):
+    def add_item(self, new_item):
         """
-        Adds a new item to our item list.  Returns a tuple containing the new
-        BL3Item object itself, and its new index in our item list.
+        Adds a new `new_item` (BL3Item object) to our item list.  Returns the item's
+        new index in our item list.
         """
+
+        # Add the item to the protobuf
+        self.save.inventory_items.append(new_item.protobuf)
+
+        # The protobuf reference that we append to the protobuf list
+        # ends up *not* being the one that's actually used when we
+        # save, so if we want to be able to alter it later (say, below
+        # when levelling up items), we have to grab a fresh reference
+        # to it.
+        new_item.protobuf = self.save.inventory_items[-1]
+
+        # Now update our internal items list and return
+        self.items.append(new_item)
+        return len(self.items)-1
+
+    def create_new_item(self, item_serial):
+        """
+        Creates a new item from the given binary `item_serial`, which can later
+        be added to our item list.
+        """
+
         # Okay, I have no idea what this pickup_order_index attribute is about, but let's
-        # make sure it's unique anyway.
+        # make sure it's unique anyway.  It might be related to ordering when picking
+        # up multiple items at once, which would probably make it more useful for auto-pick-up
+        # items like money and ammo...
         max_pickup_order = 0
         for item in self.items:
             if item.get_pickup_order_idx() > max_pickup_order:
                 max_pickup_order = item.get_pickup_order_idx()
 
-        # Create the item and add it in
-        new_item = BL3Item.create(serial_number=itemdata,
+        # Create the item and return it
+        new_item = BL3Item.create(self.datawrapper,
+                serial_number=item_serial,
                 pickup_order_idx=max_pickup_order+1,
                 is_favorite=True,
                 )
-        self.save.inventory_items.append(new_item.protobuf)
-        self.items.append(new_item)
-        return (self.items[-1], len(self.items)-1)
+        return new_item
 
-    def add_new_item_encoded(self, itemdata):
+    def create_new_item_encoded(self, item_serial_b64):
         """
-        Adds a new item to our item list, using the textual representation that we can
-        save to a file (which happens to be base64).  Returns a tuple containing the new
-        BL3Item object itself, and its new index in our item list.
+        Creates a new item from the base64-encoded (and "BL3()"-wrapped)
+        `item_serial_b64`, which can later be added to our item list.
         """
-        return self.add_new_item(BL3Item.decode_serial_base64(itemdata))
+        return self.create_new_item(BL3Item.decode_serial_base64(item_serial_b64))
+
+    def add_new_item(self, item_serial):
+        """
+        Adds a new item to our item list using the binary `item_serial`.
+        Returns a tuple containing the new BL3Item object itself, and its
+        new index in our item list.
+        """
+        new_item = self.create_new_item(item_serial)
+        return (new_item, self.add_item(new_item))
+
+    def add_new_item_encoded(self, item_serial_b64):
+        """
+        Adds a new item to our item list using the base64-encoded (and
+        "BL3()"-wrapped) `item_serial_b64`.  Returns a tuple containing the
+        new BL3Item object itself, and its new index in our item list.
+        """
+        return self.add_new_item(BL3Item.decode_serial_base64(item_serial_b64))
 
     def overwrite_item_in_slot(self, slot, itemdata):
         """
