@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # vim: set expandtab tabstop=4 shiftwidth=4:
 
 # Copyright (c) 2020 CJ Kucera (cj@apocalyptech.com)
@@ -23,8 +23,8 @@
 
 import os
 import sys
-import argparse
 import bl3save
+import argparse
 from bl3save.bl3save import BL3Save
 
 class DictAction(argparse.Action):
@@ -70,8 +70,13 @@ def main():
             """
             )
 
+    parser.add_argument('-V', '--version',
+            action='version',
+            version='BL3 CLI SaveEdit v{}'.format(bl3save.__version__),
+            )
+
     parser.add_argument('-o', '--output',
-            choices=['savegame', 'protobuf', 'items'],
+            choices=['savegame', 'protobuf', 'json', 'items'],
             default='savegame',
             help='Output file format',
             )
@@ -110,6 +115,18 @@ def main():
             help='Set the character to max level ({})'.format(bl3save.max_level),
             )
 
+    itemlevelgroup = parser.add_mutually_exclusive_group()
+
+    itemlevelgroup.add_argument('--items-to-char',
+            dest='items_to_char',
+            action='store_true',
+            help='Set all inventory items to the level of the character')
+
+    itemlevelgroup.add_argument('--item-levels',
+            dest='item_levels',
+            type=int,
+            help='Set all inventory items to the specified level')
+
     parser.add_argument('--mayhem',
             type=int,
             choices=range(5),
@@ -152,6 +169,11 @@ def main():
             help='Import items from file',
             )
 
+    parser.add_argument('--allow-fabricator',
+            dest='allow_fabricator',
+            action='store_true',
+            help='Allow importing Fabricator when importing items from file')
+
     # Positional args
     parser.add_argument('input_filename',
             help='Input filename',
@@ -178,6 +200,15 @@ def main():
     if args.level_max:
         args.level = bl3save.max_level
 
+    # Check item level.  The max storeable in the serial number is 127, but the
+    # effective limit in-game is 100, thanks to MaxGameStage attributes.  We
+    # could use `bl3save.max_level` here, too, of course, but in the event that
+    # I don't get this updated in a timely fashion, having it higher would let
+    # this util potentially continue to be able to level up gear.
+    if args.item_levels:
+        if args.item_levels < 1 or args.item_levels > 100:
+            raise argparse.ArgumentTypeError('Valid item level range is 1 through 100')
+
     # Check for overwrite warnings
     if os.path.exists(args.output_filename) and not args.force:
         sys.stdout.write('WARNING: {} already exists.  Overwrite [y/N]? '.format(args.output_filename))
@@ -212,6 +243,8 @@ def main():
         len(args.unlock) > 0,
         args.copy_nvhm,
         args.import_items,
+        args.items_to_char,
+        args.item_levels,
         ])
 
     # Make changes
@@ -336,11 +369,70 @@ def main():
                 for line in df:
                     itemline = line.strip()
                     if itemline.lower().startswith('bl3(') and itemline.endswith(')'):
-                        save.add_new_item_encoded(itemline)
+                        new_item = save.create_new_item_encoded(itemline)
+                        if not args.allow_fabricator:
+                            # Report these regardless of args.quiet
+                            if not new_item.eng_name:
+                                print('   - NOTICE: Skipping unknown item import because --allow-fabricator is not set')
+                                continue
+                            if new_item.balance_short.lower() == 'balance_eridian_fabricator':
+                                print('   - NOTICE: Skipping Fabricator import because --allow-fabricator is not set')
+                                continue
+                        save.add_item(new_item)
+                        if not args.quiet:
+                            if new_item.eng_name:
+                                print('   + {} (level {})'.format(new_item.eng_name, new_item.level))
+                            else:
+                                print('   + unknown item')
                         added_count += 1
             if not args.quiet:
                 print('   - Added Item Count: {}'.format(added_count))
 
+        # Setting item levels.  Keep in mind that we'll want to do this *after*
+        # various of the actions above.  If we've been asked to up the level of
+        # the character, we'll want items to follow suit, and if we've been asked
+        # to change the level of items, we'll want to do it after the item import.
+        if args.items_to_char or args.item_levels:
+            if args.items_to_char:
+                to_level = save.get_level()
+            else:
+                to_level = args.item_levels
+            num_items = len(save.get_items())
+            if not args.quiet:
+                if num_items == 1:
+                    plural = ''
+                else:
+                    plural = 's'
+                print(' - Updating {} item{} to level {}'.format(
+                    num_items,
+                    plural,
+                    to_level,
+                    ))
+            actually_updated = 0
+            for item in save.get_items():
+                if item.level != to_level:
+                    item.level = to_level
+                    actually_updated += 1
+            if not args.quiet:
+                remaining = num_items - actually_updated
+                if actually_updated == 1:
+                    updated_verb = 'was'
+                else:
+                    updated_verb = 'were'
+                if remaining > 0:
+                    if remaining == 1:
+                        remaining_verb = 'was'
+                    else:
+                        remaining_verb = 'were'
+                    remaining_txt = ' ({} {} already at that level)'.format(remaining, remaining_verb)
+                else:
+                    remaining_txt = ''
+                print('   - {} {} updated{}'.format(
+                    actually_updated,
+                    updated_verb,
+                    remaining_txt,
+                    ))
+
         # Copying NVHM state
         if args.copy_nvhm:
             if not args.quiet:
@@ -360,10 +452,19 @@ def main():
         save.save_protobuf_to(args.output_filename)
         if not args.quiet:
             print('Wrote protobuf to {}'.format(args.output_filename))
+    elif args.output == 'json':
+        save.save_json_to(args.output_filename)
+        if not args.quiet:
+            print('Wrote JSON to {}'.format(args.output_filename))
     elif args.output == 'items':
         with open(args.output_filename, 'w') as df:
             for item in save.get_items():
+                if item.eng_name:
+                    print('# {} (level {})'.format(item.eng_name, item.level), file=df)
+                else:
+                    print('# unknown item', file=df)
                 print(item.get_serial_base64(), file=df)
+                print('', file=df)
         if not args.quiet:
             print('Wrote {} items (in base64 format) to {}'.format(len(save.get_items()), args.output_filename))
     else:
